import tensorflow as tf
import csv
from numpy import asarray
from numpy import save
#print(tf.test.is_gpu_available())

print(tf.__version__) 
print(tf.config.list_physical_devices('GPU'))

"""
#sess = tf.Session(config=tf.ConfigProto(log_device_placement=True))
sess = tf.compat.v1.Session(config=tf.compat.v1.ConfigProto(log_device_placement=True))
print("Help is it gpu", sess)
"""

import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds
import matplotlib.pyplot as plt
import numpy as np
import time

# load data separated into training and test set dataset 
(ds_train, ds_test), ds_info = tfds.load('stanford_dogs', split=['train', 'test'], shuffle_files=True, with_info=True)



print("The data was just loaded and auto-split into the train and test sets. The images will now undergo conversion to fit the model's input format.")
#____________________________________________________________________________________________________________________________
# Resize the images to fit the input for AlexNet (227x227)

# Training Set
train_lst_imgs = []
train_lst_labels = []

print("The images will be resized now.")

kaggle_breeds = ('n02105162-malinois', 'n02094258-Norwich_terrier', 'n02102177-Welsh_springer_spaniel', 'n02086646-Blenheim_spaniel', 'n02086910-papillon', 'n02093256-Staffordshire_bullterrier', 'n02113624-toy_poodle', 'n02105056-groenendael', 'n02109961-Eskimo_dog', 'n02116738-African_hunting_dog', 'n02096177-cairn', 'n02096585-Boston_bull', 'n02100735-English_setter', 'n02102973-Irish_water_spaniel', 'n02099429-curly-coated_retriever', 'n02088364-beagle', 'n02101006-Gordon_setter', 'n02108089-boxer', 'n02097130-giant_schnauzer', 'n02112137-chow')

for train in tfds.as_numpy(ds_train):  # example is `{'image': tf.Tensor, 'label': tf.Tensor}`
    #print(list(example.keys()))
    
    #resize the image to 227 x 227 because this is what AlexNet uses
    #if you want it with padding: 
    #train_lst_imgs.append(tf.image.resize_with_pad(train["image"], target_height=227, target_width=227))
    
    #Do a subset of the first 20 dog breeds to make it run faster
    
    if train["label"] <= 20:
        train_lst_imgs.append(np.array(tf.image.resize(train["image"], (227, 227))))
        label = train["label"]
        train_lst_labels.append(label)
    
# Test Set
test_lst_imgs = []
test_lst_labels = []

for test in tfds.as_numpy(ds_test):  # example is `{'image': tf.Tensor, 'label': tf.Tensor}`
    #print(list(example.keys()))
    
    #resize the image to 227 x 227 because this is what AlexNet uses

    #Do a subset of the first 20 dog breeds to make it run faster
    if test["label"] <= 20:
        test_lst_imgs.append(np.array(tf.image.resize(test["image"], (227, 227))))
        label = test["label"]
        test_lst_labels.append(label)
    
    
    
#verify that all of the images/labels were transferred when they were resized
print("The length of the training list of resized images is:", len(train_lst_imgs))
print("The length of the training list of labels is:", len(train_lst_labels))

print("The length of the test list of resized images is:", len(test_lst_imgs))
print("The length of the test list of labels is:", len(test_lst_labels))

print(type(train_lst_imgs),type(train_lst_labels))
#____________________________________________________________________________________________________________________________

# Convert into NumPy Arrays

# this can take a while - in TF2.X, apparently it's slow to convert list of tensors to numpy.array
x_train = np.array(train_lst_imgs) #images
y_train = np.array(train_lst_labels) #labels


x_test = np.array(test_lst_imgs) #images
y_test = np.array(test_lst_labels) #labels

print("The images shape after converting into NumPy arrays is: ",x_train.shape,"\nThe labels shape after converting into NumPy arrays is: ",y_train.shape)
print("Their types are: ")
print(type(x_train),type(y_train))


#____________________________________________________________________________________________________________________________


# Normalize the data

# Training Set
xn_train = x_train.astype(np.float32)
yn_train = y_train.astype(np.int32)
#xn_train = xn_train/255
print("Training set images shape after normalization = ", xn_train.shape)


# Test Set
xn_test = x_test.astype(np.float32)
yn_test = y_test.astype(np.int32)
print("Test set images shape after normalization = ", xn_test.shape)



#____________________________________________________________________________________________________________________________


print("The model will now be defined and compiled.")

# Define the model

# Import a few specific libraries/layers
from tensorflow import keras
from tensorflow.keras import layers, models

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, MaxPooling2D, Dense, Flatten, BatchNormalization, Dropout

model=Sequential()

#1 conv layer
model.add(Conv2D(filters=96,kernel_size=(11,11),strides=(4,4),padding="valid",activation="relu",input_shape=(227,227,3)))

#1 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())

#2 conv layer
model.add(Conv2D(filters=256,kernel_size=(5,5),strides=(1,1),padding="valid",activation="relu"))

#2 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())

#3 conv layer
model.add(Conv2D(filters=384,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#4 conv layer
model.add(Conv2D(filters=384,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#5 conv layer
model.add(Conv2D(filters=256,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#3 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())


model.add(Flatten())

#1 dense layer
model.add(Dense(4096,input_shape=(227,227,3),activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#2 dense layer
model.add(Dense(4096,activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#3 dense layer
model.add(Dense(1000,activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#output layer
model.add(Dense(120,activation="softmax"))

model.summary()

#____________________________________________________________________________________________________________________________

# Compile the model

model.compile(optimizer="adam", loss="sparse_categorical_crossentropy", metrics=["accuracy"])

print("Model is compiled.")

#____________________________________________________________________________________________________________________________

print("Now the model will be trained. This may take around 20-30 minutes depending on image class size and epoch size.")
# Train the model

start = time.time()
history = model.fit(xn_train, yn_train, epochs=100)
end = time.time()
total_time = end-start

#____________________________________________________________________________________________________________________________

# Convert the time from seconds

def convert(seconds): 
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

total_time = convert(total_time)

#____________________________________________________________________________________________________________________________

# Save the epoch accuracy data

with open('output.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(model.history.history['accuracy'])

#____________________________________________________________________________________________________________________________

# Evaluate the loss and accuracy

la = []

loss, accuracy = model.evaluate(xn_test, yn_test)

print("Training Runtime: ", total_time)
print("The Test Set Loss is: ", loss, "\nThe Test Set Accuracy is: ", accuracy)

la.append(loss)
la.append(accuracy)
la.append(total_time)


#____________________________________________________________________________________________________________________________

# Predict values with the model

pred = model.predict(xn_test)

#____________________________________________________________________________________________________________________________

# Download xn_test and yn_test to create images of the predicted and actual dog breed estimates

#randomly select 9 images

images = []
actual = []
pred_label = []

for i in range(9):
    r = np.random.randint( 0, xn_test.shape[0], 1)
    images.append(xn_test[r[0]])
    actual.append(yn_test[r[0]])
    pred_label.append(yn_test[r[0]]*pred[r[0]][yn_test[r[0]]])

save('images.npy', images)

with open('pred_labels.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(pred_label)
    
with open('actual_labels.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(actual)
    
with open('test.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(la)