import tensorflow as tf
import csv
from numpy import asarray
from numpy import save
#print(tf.test.is_gpu_available())

print(tf.__version__) 
print(tf.config.list_physical_devices('GPU'))
import time

import os
import cv2
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split

#from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import BatchNormalization,Conv2D,MaxPooling2D,Dense,Flatten,Dropout

print("Loaded all libraries")

from matplotlib.image import imread
from PIL import Image
import os, cv2

random_seed = 42
fpath = './stanford-dogs-dataset/images/Images/'

categories = []

categories = os.listdir(fpath)

kaggle_breeds = ('n02105162-malinois', 'n02094258-Norwich_terrier', 'n02102177-Welsh_springer_spaniel', 'n02086646-Blenheim_spaniel', 'n02086910-papillon', 'n02093256-Staffordshire_bullterrier', 'n02113624-toy_poodle', 'n02105056-groenendael', 'n02109961-Eskimo_dog', 'n02116738-African_hunting_dog', 'n02096177-cairn', 'n02096585-Boston_bull', 'n02100735-English_setter', 'n02102973-Irish_water_spaniel', 'n02099429-curly-coated_retriever', 'n02088364-beagle', 'n02101006-Gordon_setter', 'n02108089-boxer', 'n02097130-giant_schnauzer', 'n02112137-chow')

"""
for breed in kaggle_breeds:
    categories.append(breed)
"""

print("List of categories = ",categories,"\n\nNo. of categories = ", len(categories))


def load_images_and_labels(categories):
    img_lst=[]
    labels=[]
    for index, category in enumerate(categories):
        for image_name in os.listdir(fpath+"/"+category):
            img = cv2.imread(fpath+"/"+category+"/"+image_name)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            
            img_array = Image.fromarray(img, 'RGB')
            
            #resize image to 227 x 227 because the input image resolution for AlexNet is 227 x 227
            resized_img = img_array.resize((227, 227))
            
            img_lst.append(np.array(resized_img))
            
            labels.append(index)
    return img_lst, labels

images, labels = load_images_and_labels(categories)

print("No. of images loaded = ",len(images),"\nNo. of labels loaded = ",len(labels))
print(type(images),type(labels))

images = np.array(images)
labels = np.array(labels)
print("After np:", type(images),type(labels))

#reformat images
images = images.astype(np.float32)
labels = labels.astype(np.int32)
images = images/255


#1-step in data shuffling

#get equally spaced numbers in a given range
n = np.arange(images.shape[0])
print("'n' values before shuffling = ",n)

#shuffle all the equally spaced values in list 'n'
np.random.seed(random_seed)
np.random.shuffle(n)
print("\n'n' values after shuffling = ",n)


#2-step in data shuffling

#shuffle images and corresponding labels data in both the lists
images = images[n]
labels = labels[n]

print("Images shape after shuffling = ",images.shape,"\nLabels shape after shuffling = ",labels.shape)



images = images.astype(np.float32)
labels = labels.astype(np.int32)
images = images/255
print("Images shape after normalization = ",images.shape)


x_train, x_test, y_train, y_test = train_test_split(images, labels, test_size = 0.2, random_state = random_seed)

print("x_train shape = ",x_train.shape)
print("y_train shape = ",y_train.shape)
print("\nx_test shape = ",x_test.shape)
print("y_test shape = ",y_test.shape)


print("Everything worked so far!!")


#____________________________________________________________________________________________________________________________


print("The model will now be defined and compiled.")

# Define the model

model=Sequential()

#1 conv layer
model.add(Conv2D(filters=96,kernel_size=(11,11),strides=(4,4),padding="valid",activation="relu",input_shape=(227,227,3)))

#1 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())

#2 conv layer
model.add(Conv2D(filters=256,kernel_size=(5,5),strides=(1,1),padding="valid",activation="relu"))

#2 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())

#3 conv layer
model.add(Conv2D(filters=384,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#4 conv layer
model.add(Conv2D(filters=384,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#5 conv layer
model.add(Conv2D(filters=256,kernel_size=(3,3),strides=(1,1),padding="valid",activation="relu"))

#3 max pool layer
model.add(MaxPooling2D(pool_size=(3,3),strides=(2,2)))

model.add(BatchNormalization())


model.add(Flatten())

#1 dense layer
model.add(Dense(4096,input_shape=(227,227,3),activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#2 dense layer
model.add(Dense(4096,activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#3 dense layer
model.add(Dense(1000,activation="relu"))

model.add(Dropout(0.4))

model.add(BatchNormalization())

#output layer
model.add(Dense(120,activation="softmax"))

model.summary()

#____________________________________________________________________________________________________________________________

# Compile the model

model.compile(optimizer="adam", loss="sparse_categorical_crossentropy", metrics=["accuracy"])

print("Model is compiled.")

#____________________________________________________________________________________________________________________________



print("Now the model will be trained. This may take around 3-5 minutes depending on image class size and epoch size.")
# Train the model

start = time.time()
model.fit(x_train, y_train, epochs=100)
end = time.time()
total_time = end-start

#____________________________________________________________________________________________________________________________

# Convert the time from seconds

def convert(seconds): 
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

total_time = convert(total_time)

#____________________________________________________________________________________________________________________________

# Save the epoch accuracy data

with open('output_kaggle.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(model.history.history['accuracy'])

#____________________________________________________________________________________________________________________________


# Evaluate the loss and accuracy

la_kaggle = []

#loss, accuracy = model.evaluate(xn_test, yn_test)
loss, accuracy = model.evaluate(x_test, y_test)

print("Training Runtime: ", total_time)
print("The Test Set Loss is: ", loss, "\nThe Test Set Accuracy is: ", accuracy)

la_kaggle.append(loss)
la_kaggle.append(accuracy)
la_kaggle.append(total_time)



#____________________________________________________________________________________________________________________________

# Predict values with the model

pred = model.predict(x_test)

pred.shape



#____________________________________________________________________________________________________________________________

# Download xn_test and yn_test to create images of the predicted and actual dog breed estimates

#randomly select 8 images

kaggle_images= []
kaggle_actual = []
kaggle_pred_label = []
n=0

for i in range(9):
    n+=1 
    r = np.random.randint(0, x_test.shape[0], 1)
    kaggle_images.append(x_test[r[0]]*255)
    kaggle_actual.append(y_test[r[0]])
    kaggle_pred_label.append(y_test[r[0]]*pred[r[0]][y_test[r[0]]])

save('kaggle_images.npy', kaggle_images)

with open('kaggle_pred_label.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(kaggle_pred_label)
    
with open('kaggle_actual.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(kaggle_actual)
    
with open('kaggle_test_results.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(la_kaggle)
    